;DCF77 receiver - completely rewritten version
;with output for four 7-Segment LED/VF displays
;Controller: AT90S2313 @ 4.194304 MHz
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.elektronik.de.vu
;
;VERSION HISTORY:
;  0.1	2005-06-21
;  0.2	2005-11-22
;	* added support for active-high common outputs
;  0.3	2006-12-23
;	* added support for 18:88 type display (12-hour format)
;	* disable DCF77 reception after 4 minutes if nothing was received
;	* minor bugfixes
;  0.3	2007-01-14
;	* added support for 18:88 type display (12-hour format)
;	* disable DCF77 reception after 4 minutes if nothing was received
;	* bugfixes for bugs introduced with 12-hour-format version

;microcontroller definition file
.include "../include/2313def.inc"

;DCF77 receiver
.equ	DCFPort = PORTD
.equ	DCFPin = 2
.equ	DCFInvert = 1 ;inverted output (negative pulses), most receivers do this!
.equ	DCFPullup = 1 ;internal pull-up resistor for DCF77 input

;LED/VF display
.equ	LEDSegmentPort = PORTB ;display segment outputs, LSB=a, MSB=dp
.equ	LEDCommonPort = PORTD ;display common outputs (display select)
.equ	LEDOuterLeft = 3
.equ	LEDInnerLeft = 4
.equ	LEDInnerRight = 5
.equ	LEDOuterRight = 6
.equ	LEDCommonInvert = 1 ;0 for NPN transistors
.equ	LEDSegmentInvert = 1 ;0 for NPN transistors and directly connected common-cathode displays
.equ	TwelveHour = 0 ;12 hour display (18:88)
.equ	ColonDisp = 1 ;display # which contains the colon (0 = outer left, 3 = outer right)
.equ	ColonMask = 0x80 ;bit mask for colon (0x80 for dp)

;Key for date display:
.equ	KeyPort = PORTD
.equ	KeyPin = 0
.equ	DateDisplayTime = 5 ;in seconds


;needed for lpm instruction
.def	lpmresult = R0
; ; 
;registers for DCF77 telegram
.def	dcf1 = R3
.def	dcf2 = R4
.def	dcf3 = R5
.def	dcf4 = R6
.def	dcf5 = R7

;time/date registers
.def	subsec = R8
.def	second = R9
.def	minute = R10
.def	hour = R11
.def	day = R12
.def	month = R13
.def	year = R14
.def	wkday = R15 ;(weekday, 1=MO 7=SU)

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/64-seconds since last edge
.def	flags1 = R20
.def	flags2 = R21

;additional registers for multiplexing
.def	display = R22 ;active display [0..3]
.def	common = R23 ;display common pin mask

;key registers
.def	datecnt = R24

;bit definitions for the flags1 register
.equ	ErrBitCount = 0 ;last error: wrong number of bits
.equ	ErrParity = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit = 3 ;last error: start bit was not 1
.equ	TimeOK = 4 ;time is OK
.equ	DateOK = 5 ;date is OK (reset at midnight)
.equ	StopDCF = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFPin ;last state of DCF77 input, must be the same bit as on the port!

.cseg

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

;Reset
.org $0000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xFF
	out	LEDSegmentPort-1, temp ;DDRx
	.if ( LEDCommonInvert )
		ldi	temp, 1<<LEDOuterLeft | 1<<LEDInnerLeft | 1<<LEDInnerRight | 1<<LEDOuterRight
	.elif
		ldi	temp, 0x00
	.endif
	out	LEDCommonPort-1, temp ;DDRx
	
	cbi	DCFPort-1, DCFPin ;DDRx
	.if ( DCFPullup )
		sbi	DCFPort, DCFPin ;Pullup on
	.else
		cbi	DCFPort, DCFPin ;Pullup off
	.endif
	
	cbi	KeyPort-1, KeyPin ;DDRx
	sbi	KeyPort, KeyPin ;Pullup
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
	ldi	display, 0
	ldi	datecnt, 0
	
	in	common, LEDCommonPort
	mov	temp, common
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   modifies: temp, temp2, display, common, ZL, ZH, lpmresult
;---------------------------------------

;multiplex all the time and do *all* the action while one of the displays is enabled
main_loop:
	;enable display
	out	LEDCommonPort, temp ;disable all displays
	out	LEDSegmentPort, temp2 ;output new display data
	out	LEDCommonPort, common ;enable display
	
	;check key
	cpi	datecnt, 0
	brne	key_check_end
	sbis	KeyPort-2, KeyPin ;PINx
	ldi	datecnt, DateDisplayTime*2 ;DDT seconds date display, then DDT seconds time display
key_check_end:
	
	;encode the next display (also save the value for the next display common output pin)
	inc	display
	andi	display, 0x03 ;return to display 0 after display 3	
	cpi	display, 0
	brne	encode_1
	mov	ZL, hour
	.if ( TwelveHour )
		cpi	ZL, 0x13 ;hour >= 13: subtract 12
		brlo	PC+4
		subi	ZL, 0x12
		brhc	PC+2
		subi	ZL, 0x06
		cpi	ZL, 0x00 ;hour = 0: set hour to 12
		brne	PC+2
		ldi	ZL, 0x12
	.endif
	cpi	datecnt, DateDisplayTime+1
	brlo	nodate1
	sbrc	flags1, DateOK
	mov	ZL, month
nodate1:
	cpi	ZL, 0x10 ;suppress leading zero
	brsh	PC+2
	.if ( TwelveHour )
		ldi	ZL, 0xA0 ;(replace by '-' which isn't displayed on a 18:88 display)
	.else
		ldi	ZL, 0xF0 ;(replace by space)
	.endif
	swap	ZL
	.if ( LEDCommonInvert )
		ldi	common, ~(1<<LEDOuterLeft)
	.else
		ldi	common, 1<<LEDOuterLeft
	.endif
	sbrc	flags1, TimeOK ;if no time received yet: show last received bit (0=, 1=' in 12 hour mode) or nothing (stopped)
	rjmp	encode_data
	.if ( TwelveHour )
		ldi	ZL, 14 ;,
	.else	
		ldi	ZL, 0
	.endif
	sbrc	dcf5, 7
	.if ( TwelveHour )
		ldi	ZL, 15 ;'
	.else	
		ldi	ZL, 1
	.endif
	sbrc	flags1, StopDCF
	ldi	ZL, 10 ;-

	rjmp	encode_data
encode_1:
	cpi	display, 1
	brne	encode_2
	mov	ZL, hour
	.if ( TwelveHour )
		cpi	ZL, 0x13 ;hour >= 13: subtract 12
		brlo	PC+4
		subi	ZL, 0x12
		brhc	PC+2
		subi	ZL, 0x06
		cpi	ZL, 0x00 ;hour = 0: set hour to 12
		brne	PC+2
		ldi	ZL, 0x12
	.endif
	cpi	datecnt, DateDisplayTime+1
	brlo	nodate2
	sbrc	flags1, DateOK
	mov	ZL, month
nodate2:
	.if ( LEDCommonInvert )
		ldi	common, ~(1<<LEDInnerLeft)
	.else
		ldi	common, 1<<LEDInnerLeft
	.endif
	sbrc	flags1, TimeOK ;if no time received yet: show last error or -
	rjmp	encode_data
	ldi	ZL, 10 ;-
	sbrc	flags1, ErrParity
	ldi	ZL, 13 ;P
	sbrc	flags1, ErrStartBit
	ldi	ZL, 5 ;S
	sbrc	flags1, ErrBitCount
	ldi	ZL, 12 ;C
	sbrc	flags1, ErrBitLength
	ldi	ZL, 11 ;L
	rjmp	encode_data
encode_2:
	cpi	display, 2
	brne	encode_3
	mov	ZL, minute
	cpi	datecnt, DateDisplayTime+1
	brlo	nodate3
	sbrc	flags1, DateOK
	mov	ZL, day
nodate3:
	swap	ZL
	.if ( LEDCommonInvert )
		ldi	common, ~(1<<LEDInnerRight)
	.else
		ldi	common, 1<<LEDInnerRight
	.endif
	sbrc	flags1, TimeOK ;if no time received yet: show bit count 10's
	rjmp	encode_data
	mov	temp, dcfbits
	rcall	bin2bcd
	mov	ZL, temp
	swap	ZL
	rjmp	encode_data
encode_3:
	mov	ZL, minute
	cpi	datecnt, DateDisplayTime+1
	brlo	nodate4
	sbrc	flags1, DateOK
	mov	ZL, day
nodate4:
	.if ( LEDCommonInvert )
		ldi	common, ~(1<<LEDOuterRight)
	.else
		ldi	common, 1<<LEDOuterRight
	.endif
	sbrc	flags1, TimeOK ;if no time received yet: show bit count 1's
	rjmp	encode_data
	mov	temp, dcfbits
	rcall	bin2bcd
	mov	ZL, temp
	
encode_data:
	;encode to 7-segment
	andi	ZL, 0x0F
	ldi	temp, LOW(encoder*2)
	add	ZL, temp
	ldi	temp, HIGH(encoder*2)
	adc	ZH, temp
	lpm
	mov	temp2, lpmresult
	
	;colon shows DCF77 signal (when the receiver is enabled) or is always on (time) / off (date)
	cpi	display, ColonDisp
	brne	colon_off
	sbrs	flags1, DCFdisabled
	rjmp	colon_dcf
	cpi	datecnt, DateDisplayTime+1
	brsh	colon_off
	rjmp	colon_on
colon_dcf:	
	sbrs	flags2, LastState
	rjmp	colon_off
colon_on:
	.if ( LEDSegmentInvert ) ;colon is segment d on display 0
		andi	temp2, ~(ColonMask)
	.else
		ori	temp2, ColonMask
	.endif
colon_off:

	;calculate the next display mask
	in	temp, LEDCommonPort
	.if ( LEDCommonInvert )
		.message "LED Common  Ports: INVERTED"
		ori	temp, 1<<LEDOuterLeft | 1<<LEDInnerLeft | 1<<LEDInnerRight | 1<<LEDOuterRight
		and	common, temp
	.else
		.message "LED Common  Ports: normal"
		andi	temp, ~(1<<LEDOuterLeft | 1<<LEDInnerLeft | 1<<LEDInnerRight | 1<<LEDOuterRight)
		or	common, temp
	.endif
	
	
	;wait for a while
	ldi	ZH, 4194304 / 770000 ;~1ms (for n seconds use f_ck / 770*[1/n])
	ldi	ZL, 0
multiplex_wait:
	dec	ZL
	brne	multiplex_wait ;inner loop takes 770 clock cycles to complete
	dec	ZH
	brne	multiplex_wait	
	
	rjmp	main_loop
	
;---------------------------------------
; 7-Segment encoding table
;---------------------------------------

encoder:
.if ( LEDSegmentInvert )
	.message "LED Segment Ports: INVERTED"
	;     0     1     2     3     4     5     6     7
	.db 0xC0, 0xF9, 0xA4, 0xB0, 0x99, 0x92, 0x82, 0xF8
	;     8     9     -     L     C     P
	.db 0x80, 0x90, 0xBF, 0xC7, 0xC6, 0x8C
	.if ( TwelveHour )
		;     ,     '
		.db 0xFB, 0xFD
	.else
		;   <Spc> <Spc>
		.db 0xFF, 0xFF
	.endif
.else 
	.message "LED Segment Ports: normal"
	.db 0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07
	.db 0x7F, 0x6F, 0x40, 0x38, 0x39, 0x73
	.if ( TwelveHour )
		;     ,     '
		.db 0x04, 0x02
	.else
		;   <Spc> <Spc>
		.db 0x00, 0x00
	.endif
.endif

;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5,
;             subsec, second, minute, hour, day, wkday, month, year
;---------------------------------------

ovf0int:
	;save registers
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags1, DCFdisabled
	rjmp	dcf_end
	
	inc	dcftime
	
	;compare DCF77 input to last saved state
	in	temp, DCFPort-2 ;PINx
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert )
		.message "DCF77 Signal: INVERTED"
		brne	dcf_end ;no change
	.else
		.message "DCF77 Signal: normal"
		breq	dcf_end ;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
		; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
	
	;exit if an error was detected before
	sbrc	flags1, StopDCF
	rjmp	dcf_end
	 
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brlo	dcf_evalbit
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end
	
		; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 77
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	andi	flags1, ~ErrMask & ~(1<<StopDCF)
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	breq	dcf_eval
	ori	flags1, 1<<ErrBitCount
dcf_end:

;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	inctime_end
	
	;decrement date display timer
	cpi	datecnt, 0
	breq	PC+2
	dec	datecnt
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	andi	flags1, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	
	;new day: mark date as bad until the new one was received
	clr	hour
	andi	flags1, ~(1<<DateOK)
inctime_end:

	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 reception (continued)
;---------------------------------------

		; < negative edge - cont'd >
	
	;check bit length (optional)
dcf_evalbit:
	cpi	dcftime, 5 ;bit is 0: 5 <= dcftime <= 7
	brlo	dcf_errbit
	cpi	dcftime, 8
	brlo	dcf_clearbit
	cpi	dcftime, 11 ;bit is 1: 11 <= dcftime <= 13
	brlo	dcf_errbit
	cpi	dcftime, 14
	brlo	dcf_setbit
	
	;received bit is erroneous
dcf_errbit:
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
	;save bit
dcf_clearbit:
	clc ;clear bit
	rjmp	dcf_savebit
dcf_setbit:
	sec ;set bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
	
;---------------------------------------
; evaluate DCF77 telegram
;---------------------------------------
	
dcf_eval:
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	dcf_end
	
dcf_parityerror:
	;set parity error
	ori	flags1, 1<<ErrParity
	rjmp	dcf_end

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; calculate parity and add up
;   in: temp
;   out: temp2 (add to existing value)
;---------------------------------------

parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret
	
;---------------------------------------
; increment packed bcd number
;   in: temp
;   out: temp
;---------------------------------------

bcdinc:
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;---------------------------------------
; convert binary to packed BCD
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bin2bcd:
	ldi	temp2, 0
bin2bcd_loop:
	subi	temp2, -0x10
	subi	temp, 10
	brcc	bin2bcd_loop
	subi	temp2, 0x10
	subi	temp, -10
	add	temp, temp2
	ret